# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Miscellaneous types."""

__all__ = ["JSONArray", "JSONObject", "JSONValue"]

from typing import Dict, Sequence, Union

#
# Types that can be represented in JSON.
#

JSONValue = Union[str, int, float, bool, None, "JSONArray", "JSONObject"]
JSONArray = Sequence[JSONValue]
JSONObject = Dict[str, JSONValue]
